
from aperiot.lexer import *

imports = ['aperiot.lexer']

useindents = True
usenewlines = True

# Terminal symbols

indent  = Indent
dedent  = Dedent
newline = Newline
importkwd      = Keyword('import', 'importkwd')
useindentskwd  = Keyword('useindents', 'useindentskwd')
usenewlineskwd = Keyword('usenewlines', 'usenewlineskwd')
keywordskwd    = Keyword('keywords', 'keywordskwd')
operatorskwd   = Keyword('operators', 'operatorskwd')
separatorskwd  = Keyword('separators', 'separatorskwd')
bracketskwd    = Keyword('brackets', 'bracketskwd')
identifierskwd = Keyword('identifiers', 'identifierskwd')
numberskwd     = Keyword('numbers', 'numberskwd')
stringskwd     = Keyword('strings', 'stringskwd')
startkwd       = Keyword('start', 'startkwd')
ruleskwd       = Keyword('rules', 'ruleskwd')
emptykwd       = Keyword('empty', 'emptykwd')
arrow = Operator('->', 'arrow')
colon = Separator(':', 'colon')
bar   = Separator('|', 'bar')
dot   = Separator('.', 'dot')
nonterminal = Identifier
name        = Identifier
terminal = StringLiteral
action   = StringLiteral


# Non-terminal symbols

KEYWORD              = 'KEYWORD'
KEYWORDS_SECTION     = 'KEYWORDS_SECTION'
SYMBOL               = 'SYMBOL'
NUMBER               = 'NUMBER'
BRACKETS_SECTION     = 'BRACKETS_SECTION'
NUMBERS_SECTION      = 'NUMBERS_SECTION'
LIST_OF_ALTERNATIVES = 'LIST_OF_ALTERNATIVES'
IDENTIFIER_LINE      = 'IDENTIFIER_LINE'
OPERATORS_SECTION    = 'OPERATORS_SECTION'
SEPARATORS_LINES     = 'SEPARATORS_LINES'
TERMINAL             = 'TERMINAL'
WORD                 = 'WORD'
OPERATORS_LINES      = 'OPERATORS_LINES'
STRINGS_SECTION      = 'STRINGS_SECTION'
KEYWORD_LINE         = 'KEYWORD_LINE'
RULES_SECTION        = 'RULES_SECTION'
KEYWORD_LINES        = 'KEYWORD_LINES'
IMPORT               = 'IMPORT'
SEPARATORS_SECTION   = 'SEPARATORS_SECTION'
IDENTIFIER           = 'IDENTIFIER'
RULE_LIST            = 'RULE_LIST'
BRACKETS_LINES       = 'BRACKETS_LINES'
RULE                 = 'RULE'
GRAMMAR              = 'GRAMMAR'
STRING               = 'STRING'
STRINGS_LINES        = 'STRINGS_LINES'
USEINDENTS           = 'USEINDENTS'
TERMINALS_SECTIONS   = 'TERMINALS_SECTIONS'
START_SECTION        = 'START_SECTION'
TERMINALS_SECTION    = 'TERMINALS_SECTION'
BRACKETS_LINE        = 'BRACKETS_LINE'
RULE_TABLE           = 'RULE_TABLE'
OPERATORS_LINE       = 'OPERATORS_LINE'
IDENTIFIERS_SECTION  = 'IDENTIFIERS_SECTION'
USENEWLINES          = 'USENEWLINES'
ALL                  = 'ALL'
IMPORTS_LINES        = 'IMPORTS_LINES'
SYMBOL_LIST          = 'SYMBOL_LIST'
RHS                  = 'RHS'
SEPARATORS_LINE      = 'SEPARATORS_LINE'
NUMBERS_LINES        = 'NUMBERS_LINES'
EMPTY_LINES          = 'EMPTY_LINES'
IMPORTS_SECTION      = 'IMPORTS_SECTION'
IMPORTS_LINE         = 'IMPORTS_LINE'
IDENTIFIER_LINES     = 'IDENTIFIER_LINES'
NUMBERS_LINE         = 'NUMBERS_LINE'
ALTERNATIVE          = 'ALTERNATIVE'
NONTERMINAL          = 'NONTERMINAL'
STRINGS_LINE         = 'STRINGS_LINE'


# Start symbol

start = 'ALL'

# Rules

rules = \
    {
        KEYWORD: 
            [
                ([terminal], 'Keyword($1.string)'), 
                ([name, terminal], 'Keyword($2.string, $1.string)'), 
            ],
        KEYWORDS_SECTION: 
            [
                ([keywordskwd, newline, indent, KEYWORD_LINES, dedent], '$4'), 
            ],
        SYMBOL: 
            [
                ([TERMINAL], '$1'), 
                ([NONTERMINAL], '$1'), 
            ],
        NUMBER: 
            [
                ([name], 'Token($1.string)'), 
            ],
        BRACKETS_SECTION: 
            [
                ([bracketskwd, newline, indent, BRACKETS_LINES, dedent], '$4'), 
            ],
        NUMBERS_SECTION: 
            [
                ([numberskwd, newline, indent, NUMBERS_LINES, dedent], '$4'), 
            ],
        LIST_OF_ALTERNATIVES: 
            [
                ([bar, ALTERNATIVE, newline], '[$2]'), 
                ([bar, ALTERNATIVE, newline, LIST_OF_ALTERNATIVES], '[$2] + $4'), 
            ],
        IDENTIFIER_LINE: 
            [
                ([IDENTIFIER, newline], '[$1]'), 
                ([IDENTIFIER, IDENTIFIER_LINE], '[$1] + $2'), 
            ],
        OPERATORS_SECTION: 
            [
                ([operatorskwd, newline, indent, OPERATORS_LINES, dedent], '$4'), 
            ],
        SEPARATORS_LINES: 
            [
                ([SEPARATORS_LINE], '[$1]'), 
                ([SEPARATORS_LINE, SEPARATORS_LINES], '[$1] + $2'), 
            ],
        TERMINAL: 
            [
                ([terminal], 'Token($1.string)'), 
            ],
        WORD: 
            [
                ([SYMBOL_LIST], '$1'), 
                ([emptykwd], '[]'), 
            ],
        OPERATORS_LINES: 
            [
                ([OPERATORS_LINE], '[$1]'), 
                ([OPERATORS_LINE, OPERATORS_LINES], '[$1] + $2'), 
            ],
        STRINGS_SECTION: 
            [
                ([stringskwd, newline, indent, STRINGS_LINES, dedent], '$4'), 
            ],
        KEYWORD_LINE: 
            [
                ([KEYWORD, newline], '[$1]'), 
                ([KEYWORD, KEYWORD_LINE], '[$1] + $2'), 
            ],
        RULES_SECTION: 
            [
                ([ruleskwd, EMPTY_LINES, RULE_TABLE], '$3'), 
            ],
        KEYWORD_LINES: 
            [
                ([KEYWORD_LINE], '$1'), 
                ([KEYWORD_LINE, KEYWORD_LINES], '$1 + $2'), 
            ],
        IMPORT: 
            [
                ([name], '$1.string'), 
                ([name, dot, IMPORT], '$1.string + \'.\' + $3'), 
            ],
        SEPARATORS_SECTION: 
            [
                ([separatorskwd, newline, indent, SEPARATORS_LINES, dedent], '$4'), 
            ],
        IDENTIFIER: 
            [
                ([name], 'Identifier($1.string)'), 
            ],
        RULE_LIST: 
            [
                ([RULE], '[$1]'), 
                ([RULE, RULE_LIST], '[$1] + $2'), 
            ],
        BRACKETS_LINES: 
            [
                ([BRACKETS_LINE], '[$1]'), 
                ([BRACKETS_LINE, BRACKETS_LINES], '[$1] + $2'), 
            ],
        RULE: 
            [
                ([NONTERMINAL, arrow, RHS], '($1, $3)'), 
            ],
        GRAMMAR: 
            [
                ([IMPORTS_SECTION, USEINDENTS, USENEWLINES, TERMINALS_SECTIONS, START_SECTION, RULES_SECTION], '($1, $2, $3, dict($4), $5, $6)'), 
            ],
        STRING: 
            [
                ([name], 'Token($1.string)'), 
            ],
        STRINGS_LINES: 
            [
                ([STRINGS_LINE], '$1'), 
                ([STRINGS_LINE, STRINGS_LINES], '$1 + $2'), 
            ],
        USEINDENTS: 
            [
                ([], 'False'), 
                ([useindentskwd, newline], 'True'), 
            ],
        TERMINALS_SECTIONS: 
            [
                ([], '[]'), 
                ([TERMINALS_SECTION, TERMINALS_SECTIONS], '[$1] + $2'), 
            ],
        START_SECTION: 
            [
                ([startkwd, newline, indent, NONTERMINAL, newline, dedent], '$4'), 
            ],
        TERMINALS_SECTION: 
            [
                ([KEYWORDS_SECTION], '(\'keywords\', $1)'), 
                ([OPERATORS_SECTION], '(\'operators\', $1)'), 
                ([SEPARATORS_SECTION], '(\'separators\', $1)'), 
                ([BRACKETS_SECTION], '(\'brackets\', $1)'), 
                ([IDENTIFIERS_SECTION], '(\'identifiers\', $1)'), 
                ([NUMBERS_SECTION], '(\'numbers\', $1)'), 
                ([STRINGS_SECTION], '(\'strings\', $1)'), 
            ],
        BRACKETS_LINE: 
            [
                ([name, terminal, newline], 'Bracket($2.string, $1.string)'), 
            ],
        RULE_TABLE: 
            [
                ([RULE_LIST], 'dict($1)'), 
            ],
        OPERATORS_LINE: 
            [
                ([name, terminal, newline], 'Operator($2.string, $1.string)'), 
            ],
        IDENTIFIERS_SECTION: 
            [
                ([identifierskwd, newline, indent, IDENTIFIER_LINES, dedent], '$4'), 
            ],
        USENEWLINES: 
            [
                ([], 'False'), 
                ([usenewlineskwd, newline], 'True'), 
            ],
        ALL: 
            [
                ([GRAMMAR], '$1'), 
            ],
        IMPORTS_LINES: 
            [
                ([IMPORTS_LINE], '$1'), 
                ([IMPORTS_LINE, IMPORTS_LINES], '$1 + $2'), 
            ],
        SYMBOL_LIST: 
            [
                ([SYMBOL, SYMBOL_LIST], '[$1] + $2'), 
                ([SYMBOL], '[$1]'), 
            ],
        RHS: 
            [
                ([ALTERNATIVE, newline], '[$1]'), 
                ([ALTERNATIVE, newline, indent, LIST_OF_ALTERNATIVES, dedent], '[$1] + $4'), 
            ],
        SEPARATORS_LINE: 
            [
                ([name, terminal, newline], 'Separator($2.string, $1.string)'), 
            ],
        NUMBERS_LINES: 
            [
                ([NUMBERS_LINE], '$1'), 
                ([NUMBERS_LINE, NUMBERS_LINES], '$1 + $2'), 
            ],
        EMPTY_LINES: 
            [
                ([newline], 'None'), 
                ([newline, EMPTY_LINES], 'None'), 
            ],
        IMPORTS_SECTION: 
            [
                ([], '[]'), 
                ([importkwd, newline, indent, IMPORTS_LINES, dedent], '$4'), 
            ],
        IMPORTS_LINE: 
            [
                ([IMPORT, newline], '[$1]'), 
                ([IMPORT, IMPORTS_LINE], '[$1] + $2'), 
            ],
        IDENTIFIER_LINES: 
            [
                ([IDENTIFIER_LINE], '$1'), 
                ([IDENTIFIER_LINE, IDENTIFIER_LINES], '$1 + $2'), 
            ],
        NUMBERS_LINE: 
            [
                ([NUMBER, newline], '[$1]'), 
                ([NUMBER, NUMBERS_LINE], '[$1] + $2'), 
            ],
        ALTERNATIVE: 
            [
                ([WORD], '($1, None)'), 
                ([WORD, colon, action], '($1, $3.value)'), 
            ],
        NONTERMINAL: 
            [
                ([nonterminal], '$1.string'), 
            ],
        STRINGS_LINE: 
            [
                ([STRING, newline], '[$1]'), 
                ([STRING, STRINGS_LINE], '[$1] + $2'), 
            ],
    }

