"""
Module: metacfg_wat
Author: Ernesto Posse
Created on: Jul 4, 2006
Last update: Jul 31, 2006

Description:  
  This module contains the meta-cfg (the CFG of CFGs) with action templates, to compile into the Python dictionary form of CFGs recognized by the CFG class.
"""

from aperiot.lexer import *

imports = ['aperiot.lexer']

useindents = True
usenewlines = True

#Tokens

newline        = Newline
indent         = Indent
dedent         = Dedent
name           = Identifier
terminal       = StringLiteral
nonterminal    = Identifier
action         = StringLiteral
arrow          = Operator('->', 'arrow')
bar            = Separator('|', 'bar')
colon          = Separator(':', 'colon')
dot            = Separator('.', 'dot')
emptystring    = Keyword('empty')
startkwd       = Keyword('start')
ruleskwd       = Keyword('rules')
importkwd      = Keyword('import')
useindentskwd  = Keyword('useindents')
usenewlineskwd = Keyword('usenewlines')
keywordskwd    = Keyword('keywords')
operatorskwd   = Keyword('operators')
separatorskwd  = Keyword('separators')
bracketskwd    = Keyword('brackets')
identifierskwd = Keyword('identifiers')
numberskwd     = Keyword('numbers')
stringskwd     = Keyword('strings')

#Non-terminals

TERMINAL             = 'TERMINAL'
NONTERMINAL          = 'NONTERMINAL'
SYMBOL               = 'SYMBOL'
SYMBOL_LIST          = 'SYMBOL_LIST'
WORD                 = 'WORD'
ALTERNATIVE          = 'ALTERNATIVE'
LIST_OF_ALTERNATIVES = 'LIST_OF_ALTERNATIVES'
RHS                  = 'RHS'
RULE                 = 'RULE'
RULE_LIST            = 'RULE_LIST'
RULE_TABLE           = 'RULE_TABLE'
GRAMMAR              = 'GRAMMAR'
EMPTY_LINES          = 'EMPTY_LINES'
IMPORTS_SECTION      = 'IMPORTS_SECTION'
IMPORTS_LINES        = 'IMPORTS_LINES'
IMPORTS_LINE         = 'IMPORTS_LINE'
IMPORT               = 'IMPORT'
USEINDENTS           = 'USEINDENTS'
USENEWLINES          = 'USENEWLINES'
KEYWORDS_SECTION     = 'KEYWORDS_SECTION'
KEYWORD_LINES        = 'KEYWORD_LINES'
KEYWORD_LINE         = 'KEYWORD_LINE'
KEYWORD              = 'KEYWORD'
OPERATORS_SECTION    = 'OPERATORS_SECTION'
OPERATORS_LINES      = 'OPERATORS_LINES'
OPERATORS_LINE       = 'OPERATORS_LINE'
SEPARATORS_SECTION   = 'SEPARATORS_SECTION'
SEPARATORS_LINES     = 'SEPARATORS_LINES'
SEPARATORS_LINE      = 'SEPARATORS_LINE'
BRACKETS_SECTION     = 'BRACKETS_SECTION'
BRACKETS_LINES       = 'BRACKETS_LINES'
BRACKETS_LINE        = 'BRACKETS_LINE'
IDENTIFIERS_SECTION  = 'IDENTIFIERS_SECTION'
IDENTIFIER_LINES     = 'IDENTIFIERS_LINES'
IDENTIFIER_LINE      = 'IDENTIFIERS_LINE'
IDENTIFIER           = 'IDENTIFIER'
NUMBERS_SECTION      = 'NUMBERS_SECTION'
NUMBERS_LINES        = 'NUMBERS_LINES'
NUMBERS_LINE         = 'NUMBERS_LINE'
NUMBER               = 'NUMBER'
STRINGS_SECTION      = 'STRINGS_SECTION'
STRINGS_LINES        = 'STRINGS_LINES'
STRINGS_LINE         = 'STRINGS_LINE'
STRING               = 'STRING'
RULES_SECTION        = 'RULES_SECTION'
TERMINALS_SECTIONS   = 'TERMINALS_SECTIONS'
TERMINALS_SECTION    = 'TERMINALS_SECTION'
START_SECTION        = 'START_SECTION'
ALL                  = 'ALL'

start = ALL

rules = \
    {
        ALL:
            [
                ([GRAMMAR], '$1')
            ],
        GRAMMAR:
            [
                ([IMPORTS_SECTION, USEINDENTS, USENEWLINES, TERMINALS_SECTIONS, START_SECTION, RULES_SECTION], '($1,$2,$3,dict($4),$5,$6)')
            ],
        RULES_SECTION:
            [
                ([ruleskwd, EMPTY_LINES, RULE_TABLE], '$3')
            ],
        START_SECTION:
            [
                ([startkwd, newline, indent, NONTERMINAL, newline, dedent], '$4')
            ],
        TERMINALS_SECTIONS:
            [
                ([], '[]'),
                ([TERMINALS_SECTION, TERMINALS_SECTIONS], '[$1] + $2')
            ],
        TERMINALS_SECTION:
            [
                ([KEYWORDS_SECTION], '("keywords", $1)'),
                ([OPERATORS_SECTION], '("operators", $1)'),
                ([SEPARATORS_SECTION], '("separators", $1)'),
                ([BRACKETS_SECTION], '("brackets", $1)'),
                ([IDENTIFIERS_SECTION], '("identifiers", $1)'),
                ([NUMBERS_SECTION], '("numbers", $1)'),
                ([STRINGS_SECTION], '("strings", $1)')
            ],
        IMPORTS_SECTION:
            [
                ([], '[]'),
                ([importkwd, newline, indent, IMPORTS_LINES, dedent], '$4')
            ],
        IMPORTS_LINES:
            [
                ([IMPORTS_LINE], '$1'),
                ([IMPORTS_LINE, IMPORTS_LINES], '$1+$2'),
            ],
        IMPORTS_LINE:
            [
                ([IMPORT,newline], '[$1]'),
                ([IMPORT, IMPORTS_LINE], '[$1]+$2')
            ],
        IMPORT:
            [
                ([name], '$1.string'),
                ([name, dot, IMPORT], '$1.string + \".\" + $3')
            ],
        USEINDENTS:
            [
                ([], 'False'),
                ([useindentskwd, newline], 'True')
            ],
        USENEWLINES:
            [
                ([], 'False'),
                ([usenewlineskwd, newline], 'True')
            ],
        KEYWORDS_SECTION:
            [
                #~ ([], '[]'),
                ([keywordskwd, newline, indent, KEYWORD_LINES, dedent], '$4')
            ],
        KEYWORD_LINES:
            [
                ([KEYWORD_LINE], '$1'),
                ([KEYWORD_LINE, KEYWORD_LINES], '$1 + $2')
            ],
        KEYWORD_LINE:
            [
                ([KEYWORD, newline], '[$1]'),
                ([KEYWORD, KEYWORD_LINE], '[$1] + $2')
            ],
        KEYWORD:
            [
                ([terminal], 'Keyword($1.string)'),
                ([name, terminal], 'Keyword($2.string, $1.string)')
            ],
        OPERATORS_SECTION:
            [
                #~ ([], '[]'),
                ([operatorskwd, newline, indent, OPERATORS_LINES, dedent], '$4')
            ],
        OPERATORS_LINES:
            [
                ([OPERATORS_LINE], '[$1]'),
                ([OPERATORS_LINE, OPERATORS_LINES], '[$1]+$2'),
            ],
        OPERATORS_LINE:
            [
                ([name,terminal,newline], 'Operator($2.string, $1.string)')
            ],
        SEPARATORS_SECTION:
            [
                #~ ([], '[]'),
                ([separatorskwd, newline, indent, SEPARATORS_LINES, dedent], '$4')
            ],
        SEPARATORS_LINES:
            [
                ([SEPARATORS_LINE], '[$1]'),
                ([SEPARATORS_LINE, SEPARATORS_LINES], '[$1]+$2'),
            ],
        SEPARATORS_LINE:
            [
                ([name,terminal,newline], 'Separator($2.string, $1.string)')
            ],
        IDENTIFIERS_SECTION:
            [
                #~ ([], '[]'),
                ([identifierskwd, newline, indent, IDENTIFIER_LINES, dedent], '$4')
            ],
        IDENTIFIER_LINES:
            [
                ([IDENTIFIER_LINE], '$1'),
                ([IDENTIFIER_LINE, IDENTIFIER_LINES], '$1 + $2')
            ],
        IDENTIFIER_LINE:
            [
                ([IDENTIFIER, newline], '[$1]'),
                ([IDENTIFIER, IDENTIFIER_LINE], '[$1] + $2')
            ],
        IDENTIFIER:
            [
                ([name], 'Identifier($1.string)'),
            ],
        NUMBERS_SECTION:
            [
                #~ ([], '[]'),
                ([numberskwd, newline, indent, NUMBERS_LINES, dedent], '$4')
            ],
        NUMBERS_LINES:
            [
                ([NUMBERS_LINE], '$1'),
                ([NUMBERS_LINE, NUMBERS_LINES], '$1 + $2')
            ],
        NUMBERS_LINE:
            [
                ([NUMBER, newline], '[$1]'),
                ([NUMBER, NUMBERS_LINE], '[$1] + $2')
            ],
        NUMBER:
            [
                ([name], 'Token($1.string)')
            ],
        STRINGS_SECTION:
            [
                #~ ([], '[]'),
                ([stringskwd, newline, indent, STRINGS_LINES, dedent], '$4')
            ],
        STRINGS_LINES:
            [
                ([STRINGS_LINE], '$1'),
                ([STRINGS_LINE, STRINGS_LINES], '$1 + $2')
            ],
        STRINGS_LINE:
            [
                ([STRING, newline], '[$1]'),
                ([STRING, STRINGS_LINE], '[$1] + $2')
            ],
        STRING:
            [
                ([name], 'Token($1.string)')
            ],
        BRACKETS_SECTION:
            [
                #~ ([], '[]'),
                ([bracketskwd, newline, indent, BRACKETS_LINES, dedent], '$4')
            ],
        BRACKETS_LINES:
            [
                ([BRACKETS_LINE], '[$1]'),
                ([BRACKETS_LINE, BRACKETS_LINES], '[$1]+$2'),
            ],
        BRACKETS_LINE:
            [
                ([name,terminal,newline], 'Bracket($2.string, $1.string)')
            ],
        EMPTY_LINES:
            [
                ([newline], 'None'),
                ([newline, EMPTY_LINES], 'None')
            ],
        RULE_TABLE:
            [
                ([RULE_LIST], 'dict($1)')
            ],
        RULE_LIST:
            [
                ([RULE], '[$1]'),
                ([RULE, RULE_LIST], '[$1]+$2')
            ],
        RULE:
            [
                ([NONTERMINAL, arrow, RHS], '($1, $3)')
            ],
        RHS:
            [
                ([ALTERNATIVE, newline], '[$1]'),
                ([ALTERNATIVE, newline, indent, LIST_OF_ALTERNATIVES, dedent], '[$1]+$4')
            ],
        LIST_OF_ALTERNATIVES:
            [
                ([bar, ALTERNATIVE, newline], '[$2]'),
                ([bar, ALTERNATIVE, newline, LIST_OF_ALTERNATIVES], '[$2]+$4')
            ],
        ALTERNATIVE:
            [
                ([WORD], '($1,None)'),
                ([WORD, colon, action], '($1,$3.value)')
            ],
        WORD:
            [
                ([SYMBOL_LIST], '$1'),
                ([emptystring], '[]')
            ],
        SYMBOL_LIST:
            [
                ([SYMBOL, SYMBOL_LIST], '[$1]+$2'),
                ([SYMBOL], '[$1]')
            ],
        SYMBOL:
            [
                ([TERMINAL], '$1'),
                ([NONTERMINAL], '$1')
            ],
        TERMINAL:
            [
                ([terminal], 'Token($1.string)')
            ],
        NONTERMINAL:
            [
                ([nonterminal], '$1.string')
            ]
    }

