"""
Project: aspa 
Module: environments
Author: Ernesto Posse
Created on: Jun 9, 2006
Last modified: Jun 9, 2006

Description:
  This provides the definitions for name environments for a language with lexical scoping.
"""

from utils import *

class Frame:
    
    def __init__(self, dictionary=None):
        if dictionary == None:
            dictionary = {}
        self.table = dict(dictionary)
        
    def __call__(self, key):
        return self.table[key]
        
    def __getitem__(self, key):
        return self.table[key]
        
    def __setitem__(self, key, value):
        self.table[key] = value
    
    def add_binding(self, key, value):
        self.table[key] = value
    
    def __str__(self):
        if len(self.table) == 0:
            s = 'empty frame\n'
        else:
            longest = max(len(key) for key in self.table)
            s = 'frame\n'
            for key in self.table:
                s = s + '  ' + fit(str(key), longest) + ': ' + str(self.table[key]) + '\n'
        return s


class Environment:
    
    def __init__(self, frame=None, base_env=None):
        if frame is None:
            self.frame = Frame()
        else:
            self.frame = frame
        self.base_env = base_env
    
    def __getitem__(self, key):
        try:
            return self.frame[key]
        except KeyError:
            if self.base_env is None:
                raise EnvironmentException(key)
            return self.base_env[key]
    
    def __setitem__(self, key, value):
        try:
            current_value = self.frame[key]
            self.frame[key] = value
        except KeyError:
            if self.base_env is None:
                raise EnvironmentException(key)
            self.base_env[key] = value
    
    def add_binding(self, name, value):
        self.frame.add_binding(name, value)
    
    def extend(self, dictionary):
        return Environment(Frame(dictionary), self)
    
    def __str__(self):
        result = '------------------------\n'
        if self.frame == None:
            return result + 'empty\n' + result
        result = result + str(self.frame) 
        if self.base_env == None:
            result = result + '------------------------\n' 
            result = result + 'bottom\n'
            result = result + '------------------------\n' 
        else:
            result = result + str(self.base_env)
        return result

Env = Environment


class EnvironmentException(Exception):
    
    def __init__(self, var=''):
        self.var = var
    
    def __str__(self):
        return "variable '"+str(self.var)+"' undefined in the current environment"
        