"""
Module: plambda-ast
Author: Ernesto Posse
Date: July 15, 2006
Last modified: Jul 16, 2006

Description:
    Abstract syntax for plambda, a polyadic lambda-calcululs
"""

from environments import *
from utils import *

class LambdaTerm:
    
    reductions = 0
    
    """
    Public instance methods
    """

    # Public instance accessors, selectors and extractors

    def execute(self, symbol_table={}):
        python_code = str(self)
        return eval(python_code, __builtins__, symbol_table)
        #~ env = Env(Frame(symbol_table))
        #~ return self.evaluate(env)
        
    def msubst(self, mapping, symbol_table={}):
        t = self
        if type(mapping) is dict:
            mapping = mapping.items()
        for var, term in mapping:
            t = t.subst(var, term)
        return t
        

    def normalize(self, symbol_table={}, l=0):
        #~ print ' '*l+'normalization #'+str(LambdaTerm.reductions)
        LambdaTerm.reductions += 1
        #~ print ' '*l+"I am a..."+self.__class__.__name__
        #~ print ' '*l+"term to normalize: ",str(self)
        #~ print ' '*l+'\nin environment:\n'
        #~ print_dict(symbol_table,l)
        #~ print
        t = self
        try:
            t = self.__rec_normalize(symbol_table, l)
        except KeyError, e:
            #~ print ' '*l+"key error:",e
            pass
        #~ print ' '*l+'stop--------------------------------'
        #~ print ' '*l+"normalized term:", str(t)
        #~ print ' '*l+"...which is a... ",t.__class__.__name__
        #~ if isinstance(t, App):
            #~ print ' '*l+"...and its function is a...",t.function.__class__.__name__
            #~ print ' '*l+"...and its arguments are ...",[arg.__class__.__name__ for arg in t.args]
        #~ elif isinstance(t, Abs):
            #~ print ' '*l+"...and its body is a...",t.body.__class__.__name__
        #~ print 
        return t
    

    """
    Private instance methods
    """
    
    # Private accessors, selectors and extractors
       
    
    # Private mutators
    
    def __rec_normalize(self, symbol_table={}, l=0):
        t = self
        #~ print ' '*l+'rec-------------------------------'
        #~ print ' '*l+'term:'
        #~ print ' '*l+'  '+str(t)
        if isinstance(t, Constant):
            #~ print ' '*l+'..const'
            pass
        elif isinstance(t, Variable):
            #~ print ' '*l+'..var'
            if t.name in symbol_table:
                t = symbol_table[t.name].term
                #~ print ' '*l+'  ---> ',t
                t = t.__rec_normalize(symbol_table, l+2)
        elif isinstance(t, LambdaAbstraction):
            #~ print ' '*l+'..abs'
            new_body = t.body.__rec_normalize(symbol_table, l+2)
            t = LambdaAbstraction(t.params, new_body)
        elif isinstance(t, Application):
            #~ print ' '*l+'..app'
            new_func = t.function.__rec_normalize(symbol_table, l+4)
            #~ print ' '*l+'..normalized function:'
            #~ print indent(str(new_func),l+6)
            #~ print ' '*l+'..normaling args...'
            new_args = [arg.__rec_normalize(symbol_table, l+4) for arg in t.args]
            #~ print ' '*l+'..normalized args:'
            #~ i=0
            #~ for arg in new_args:
                #~ i+=1
                #~ print ' '*l+'....arg '+str(i)
                #~ print indent(str(arg), l+4)
            #~ print ' '*l+'..term before replacements '
            #~ print ' '*l+'  '+str(t)
            t = Application(new_func, new_args)
            #~ print ' '*l+'..term before reduction '
            #~ print ' '*l+'  '+str(t)
            if isinstance(new_func, LambdaAbstraction) and not isinstance(new_func.body, Constant):
                t = t.beta_reduce()
                #~ print ' '*l+'..after beta red'
                #~ print ' '*l+'  '+str(t)
                t = t.__rec_normalize(symbol_table, l+2)
        #~ print ' '*l+'normalized term:'
        #~ print ' '*l+'  '+str(t)
        return t

    """
    Old methods
    """
    
    def deep_reduce(self, symbol_table={}, l=0):
        #~ print ' '*l+'reduction #'+str(LambdaTerm.reductions)
        LambdaTerm.reductions += 1
        #~ print ' '*l+"I am a..."+self.__class__.__name__
        #~ print ' '*l+"term to reduce: ",str(self)
        #~ print ' '*l+'\nin environment:\n'
        #~ print_dict(symbol_table,l)
        #~ print
        t = self
        try:
            t = self.__rec_deep_reduce(symbol_table, l)
        except KeyError, e:
            #~ print ' '*l+"key error:",e
            pass
        #~ print ' '*l+'stop--------------------------------'
        #~ print ' '*l+"reduced term:", str(t)
        #~ print ' '*l+"...which is a... ",t.__class__.__name__
        #~ if isinstance(t, App):
            #~ print ' '*l+"...and its function is a...",t.function.__class__.__name__
            #~ print ' '*l+"...and its arguments are ...",[arg.__class__.__name__ for arg in t.args]
        #~ elif isinstance(t, Abs):
            #~ print ' '*l+"...and its body is a...",t.body.__class__.__name__
        #~ print 
        return t
    
    def deep_msubst(self, mapping, symbol_table={}):
        t = self
        for var, term in mapping:
            t = t.deep_subst(var, term)
        return t
    
    def deep_subst(self, var, term):
        return self.subst(var, term)
        
    def old_eval(self, symbol_table={}):
        #~ print 'evaluation #'+str(LambdaTerm.reductions)
        LambdaTerm.reductions += 1
        #~ print str(self)
        #~ print '\nin environment:\n'
        #~ print_dict(symbol_table)
        #~ print
        t = self
        while isinstance(t, Application):
            #~ print 'iter-------------------------------'
            #~ print 'term:'
            #~ print str(t)
            if isinstance(t.function, Constant):
                #~ print 'const:',str(t.function)
                break
            elif isinstance(t.function, Variable):
                #~ print 'var:',str(t.function)
                #~ print '---> ',symbol_table[t.function.name].term
                t = Application(symbol_table[t.function.name].term, t.args)
            elif isinstance(t.function, LambdaAbstraction):
                #~ print 'abs:',str(t.function)
                t = t.beta_eval(symbol_table)
            elif isinstance(t.function, Application):
                #~ print 'app:',str(t.function)
                t = Application(t.function.eval(symbol_table), t.args)
        #~ print 'stop--------------------------------'
        #~ print str(t)
        return t
        
    def eval_msubst(self, mapping, symbol_table={}):
        t = self
        for var, term in mapping:
            t = t.subst(var, term)
        return t
        
    def get_value(self, symbol_table={}):
        return self.eval(symbol_table)
        
    def step(self):
        return self.beta_eval()

    def old_eval(self):
        #~ old = None
        #~ new = self
        #~ while new != old:
            #~ old = new
            #~ new = self.beta_eval()
        #~ return new
        t = self
        while isinstance(t, Application):
            t = t.step()
        return t
    
    def __rec_deep_reduce(self, symbol_table={}, l=0):
        t = self
        #~ print ' '*l+'rec-------------------------------'
        #~ print ' '*l+'term:'
        #~ print ' '*l+'  '+str(t)
        if isinstance(t, Constant):
            #~ print ' '*l+'..const'
            pass
        elif isinstance(t, Variable):
            #~ print ' '*l+'..var'
            if t.name in symbol_table:
                t = symbol_table[t.name].term
                #~ print ' '*l+'  ---> ',t
                t = t.__rec_deep_reduce(symbol_table, l+2)
        elif isinstance(t, LambdaAbstraction):
            #~ print ' '*l+'..abs'
            new_body = t.body.__rec_deep_reduce(symbol_table, l+2)
            t = LambdaAbstraction(t.params, new_body)
        elif isinstance(t, Application):
            #~ print ' '*l+'..app'
            new_func = t.function.__rec_deep_reduce(symbol_table, l+4)
            #~ print ' '*l+'..normalized function:'
            #~ print indent(str(new_func),l+6)
            #~ print ' '*l+'..normaling args...'
            new_args = [arg.__rec_deep_reduce(symbol_table, l+4) for arg in t.args]
            #~ print ' '*l+'..normalized args:'
            #~ i=0
            #~ for arg in new_args:
                #~ i+=1
                #~ print ' '*l+'....arg '+str(i)
                #~ print indent(str(arg), l+4)
            #~ print ' '*l+'..term before replacements '
            #~ print ' '*l+'  '+str(t)
            t = Application(new_func, new_args)
            #~ print ' '*l+'..term before reduction '
            #~ print ' '*l+'  '+str(t)
            if isinstance(new_func, LambdaAbstraction):# and not isinstance(new_func.body, Constant):
                t = t.beta_deep_reduce(symbol_table)
                t = t.__rec_deep_reduce(symbol_table, l+2)
        #~ print ' '*l+'normalized term:'
        #~ print ' '*l+'  '+str(t)
        return t
        
    def __rec_reduce(self, symbol_table={}, l=0):
        t = self
        while isinstance(t, Application):
            #~ print ' '*l+'iter-------------------------------'
            #~ print ' '*l+'term:'
            #~ print ' '*l+'  '+str(t)
            #~ print ' '*l+'  func:'
            if isinstance(t.function, Constant):
                #~ print ' '*l+'    const:',str(t.function)
                break
            elif isinstance(t.function, Variable):
                #~ print ' '*l+'    var:',str(t.function)
                #~ print ' '*l+'    ---> ',symbol_table[t.function.name].term
                t = Application(symbol_table[t.function.name].term, t.args)
            elif isinstance(t.function, LambdaAbstraction):
                #~ print ' '*l+'    abs:',str(t.function)
                if isinstance(t.function.body, Constant):
                    #~ print ' '*l+'      body is const'
                    break
                t = t.beta_reduce(symbol_table)
            elif isinstance(t.function, Application):
                #~ print ' '*l+'    app:',str(t.function)
                t = Application(t.function.__rec_reduce(symbol_table,l+2), t.args)
        return t
        
    def __str__(self):
        return '<plambda term>'
        
constants = []

class Constant(LambdaTerm):
    
    def __init__(self, value):
        global constants
        self.value = value
        constants += [self]
        
    def hasfree(variable):
        return False
        
    def freevars(self):
        return set()
        
    def subst(self, variable, term):
        return self
        
    def sim_subst(self, mapping):
        return self

    def evaluate(self, env,l=0):
        #~ print 'const eval:',str(self.value)
        #~ print 'const env:'
        #~ print env
        if type(self.value) is str \
        or isinstance(self.value, LatentFunction) \
        or isinstance(self.value, LambdaTerm):
            return eval(str(self.value), __builtins__, env)
        return self.value
        
    def __str__(self):
        return str(self.value)
        
    """
    Old methods
    """
    def deep_subst(self, variable, term):
        new_val = self.value.replace(variable, str(term))
        return Constant(new_val)
        
    #~ def msubst(self, mapping, symbol_table={}):
        #~ print '-- const msubst: "'+str(self.value)+'"'
        #~ print '   map:'
        #~ print_dict(dict(mapping), 5)
        #~ params = [param for param, arg in mapping]
        #~ args = [arg for param, arg in mapping]
        #~ abs = LambdaAbstraction(params, self)
        #~ return Application(abs, args)
        
    def eval_msubst(self, mapping, symbol_table={}):
        #~ print '-- const deep_msubst: "'+str(self.value)+'"'
        #~ print '   map:'
        #~ print_dict(dict(mapping), 5)
        #~ print '   args:'
        env = {}
        for var, term in mapping:
            if isinstance(term, Constant):
                #~ print '     arg:',var,'const:',term.value
                env[var] = term.value
            else:
                #~ print '     arg:',var,'non-const:',term.__class__.__name__,' ; ',term
                env[var] = term.get_value(symbol_table)
                #~ print '   back to: "'+str(self.value)+'"'
        #~ print '   actuals:'
        #~ print_dict(env, 5)
        #~ mapping = [(name, term.value) for name, term in mapping]
        #~ env = dict(mapping)
        env.update(symbol_table)
        result = eval(str(self.value), env)
        #~ print '   result:',result
        return result
        #~ return Constant(result)

        


class Variable(LambdaTerm):
    
    def __init__(self, name):
        self.name = name
        if type(name) is not str:
            raise SyntaxError('Bad variable name: "'+name+'" is not a string')
            
    def hasfree(variable):
        if variable == self.name:
            return True
        else:
            return False
            
    def freevars(self):
        return set([self.name])
        
    def subst(self, variable, term):
        if self.name == variable:
            return term
        else:
            return self
            
    def sim_subst(self, mapping):
        #~ print "sim subst: var:",[(p,str(mapping[p])) for p in mapping]
        if self.name in mapping:
            return mapping[self.name]
        else:
            return self
            
    def evaluate(self, env,l=0):
        #~ return env[self.name].evaluate(env)
        datum = env[self.name]
        return datum
        
    def __str__(self):
        return str(self.name)


class LambdaAbstraction(LambdaTerm):
    
    def __init__(self, params, body):
        self.params = params
        self.body = body
        if type(params) is not list:
            raise SyntaxError('Bad lambda abstraction: "'+str(params)+'" is not a valid parameter list.')
        for p in params:
            if type(p) is not str:
                raise SyntaxError('Bad lambda abstraction: parameter "'+str(p)+'" is not a valid parameter.')
        if not isinstance(body, LambdaTerm):
            raise SyntaxError('Bad lambda asbtraction: "'+str(body)+'" is not a valid plambda term')
            
    def hasfree(self, variable):
        if variable in self.params:
            return false
        return self.body.hasfree(variable)
        
    def freevars(self):
        return self.body.freevars() - set(self.params)
        
    def subst(self, variable, term):
        if variable in self.params or isinstance(self.body, Constant):
            return self
        else:
            common_vars = set(self.params) & term.freevars()
            if common_vars == set():
                return LambdaAbstraction(self.params, 
                                          self.body.subst(variable, term))
            else:
                new_params = []
                new_body = self.body
                for p in self.params:
                    if p in common_vars:
                        new_name = fresh_name()
                        print "new name: "+new_name
                        print "self:",str(self)
                        print "var:",variable
                        print "term:",str(term)
                        print "common_vars:",common_vars
                        new_params += [new_name]
                        new_body = new_body.subst(p, Var(new_name))
                        print "newbody:",new_body
                    else:
                        new_params += [p]
                return LambdaAbstraction(new_params,
                                          new_body.subst(variable,term))
    
    def sim_subst(self, mapping):
        #~ print "sim subst: abs:",[(p,str(mapping[p])) for p in mapping]
        param_set = set(self.params)
        domain = set(mapping)
        if domain <= param_set:
            return self
        else:
            common_vars_domain = param_set & domain
            if common_vars_domain == set():
                my_free_vars = self.body.freevars()
                captured = []
                for y in param_set:
                    for x in domain:
                        if x in my_free_vars and y in mapping[x].freevars():
                            captured += [y]
                if len(captured) == 0:
                    return LambdaAbstraction(self.params, self.body.sim_subst(mapping))
                else:
                    new_params = []
                    renaming = {}
                    for p in self.params:
                        if p in captured:
                            new_name = fresh_name()
                            new_params += [new_name]
                            renaming[p] = Var(new_name)
                        else:
                            new_params += [p]
                    new_body = self.body.sim_subst(renaming)
                    return LambdaAbstraction(new_params, new_body.sim_subst(mapping))
            else:
                restricted_substitution = dict((var, mapping[var]) 
                                               for var in mapping if var not in common_vars_domain)
                return self.sim_subst(restricted_substitution)
            
            
            
            #~ map_free_vars = set()
            #~ for var in mapping:
                #~ map_free_vars |= mapping[var].freevars()
            #~ common_vars_range = param_set & map_free_vars
            #~ if common_vars_range == set():
                #~ return LambdaAbstraction(self.params,
                                          #~ self.body.sim_subst(mapping))
            #~ else:
                #~ new_params = []
                #~ new_map = {}
                #~ for p in self.params:
                    #~ if p in common_vars_range:
                        #~ new_name = fresh_name()
                        #~ new_params += [new_name]
                        #~ new_map[p] = Var(new_name)
                    #~ else:
                        #~ new_params += [p]
                #~ new_body = self.body.sim_subst(new_map)
                #~ return LambdaAbstraction(new_params,
                                          #~ new_body.sim_subst(mapping))
    
    def evaluate(self, env,l=0):
        return LatentFunction(self, env)
        
    def __str__(self):
        params = ', '.join(self.params)
        return "lambda " + params + ": " + str(self.body)
        #~ return "Lambda(\n" \
            #~ + "  params: " + str(self.params) + "\n" \
            #~ + "  body: \n" \
            #~ + indent(str(self.body), 4) + "\n" \
            #~ + ")\n"

    """
    Old methods
    """
    def deep_subst(self, variable, term):
        if variable in self.params:
            return self
        else:
            common_vars = set(self.params) & term.freevars()
            if common_vars == set():
                return LambdaAbstraction(self.params, 
                                          self.body.deep_subst(variable, term))
            else:
                new_params = []
                new_body = self.body
                for p in self.params:
                    if p in common_vars:
                        new_name = fresh_name()
                        #~ print "new name: "+new_name
                        #~ print "self:",str(self)
                        #~ print "var:",variable
                        #~ print "term:",str(term)
                        #~ print "common_vars:",common_vars
                        new_params += [new_name]
                        new_body = new_body.deep_subst(p, Var(new_name))
                        #~ print "newbody:",new_body
                    else:
                        new_params += [p]
                return LambdaAbstraction(new_params,
                                          new_body.deep_subst(variable,term))
    

class Application(LambdaTerm):
    
    def __init__(self, function, args):
        self.function = function
        self.args = args
        if not isinstance(function, LambdaTerm):
            raise SyntaxError('Bad lambda application: "'+str(function)+'" is not a valid plambda term')
        if type(args) is not list and type(args) is not tuple:
            print str(type(args))
            raise SyntaxError('Bad lambda abstraction: "'+str(args)+'" is not a valid argument list.')
        new_args = []
        for p in args:
            if not isinstance(p, LambdaTerm) and not isinstance(p, LatentFunction):
                new_args += [Constant(p)]
                #~ raise SyntaxError('Bad lambda application: argument "'+str(p)+'" is not a valid plambda term.')
            else:
                new_args += [p]
        self.args = new_args
    
    def hasfree(self, variable):
        return self.function.hasfree(variable) or any(arg.hasfree(variable) for arg in self.args)
        
    def freevars(self):
        args_free_vars = set()
        for arg in self.args:
            args_free_vars |= arg.freevars()
        return self.function.freevars() | args_free_vars
    
    def subst(self, variable, term):
        return Application(self.function.subst(variable, term), \
                            [arg.subst(variable, term) for arg in self.args])
    
    def sim_subst(self, mapping):
        #~ print "sim subst: app:",[(p,str(mapping[p])) for p in mapping]
        return Application(self.function.sim_subst(mapping), \
                            [arg.sim_subst(mapping) for arg in self.args])
    
    def beta_reduce(self):
        new_subst = dict(zip(self.function.params, self.args))
        #~ print 'beta red: params subst=',[(p,str(new_subst[p])) for p in new_subst]
        #~ print 'beta red: self=',self
        #~ print 'beta red: function=',self.function
        #~ print 'beta red: function body=',self.function.body
        return self.function.body.sim_subst(new_subst)
        
    def evaluate(self, env, l=0):
        #~ print '\n--'+' '*l + 'evaluating application: \n'
        #~ print indent(str(self),l+2)
        #~ print ' '*l + 'function:\n'
        #~ print indent(str(self.function), l+2) +'\n'
        #~ print ' '*l + 'with arguments: \n'
        #~ for a in self.args:
            #~ print indent(str(a),l+2) +'\n'
            #~ print str(type(a))
            #~ if type(a) is InstanceType:
                #~ print a.__class__.__name__
            #~ print
        #~ print ' '*l+ 'in environment:\n'
        #~ print indent(str(env),l+2) +'\n'
        #~ print ' '*l +'operator type before eval:',str(type(self.function))
        #~ if type(self.function) is InstanceType:
            #~ print ' '*l+'operator class before eval:',self.function.__class__.__name__
            
        operator = self.function.evaluate(env,l+2)
        
        #~ print ' '*l +'operator type after eval:',str(type(operator))
        #~ if type(operator) is InstanceType:
            #~ print ' '*l+'operator class after eval:',operator.__class__.__name__
        #~ print ' '*l+'operator code after eval:'
        #~ print indent(str(operator), l+2)
        
        if isinstance(operator, LambdaAbstraction):
            operator = operator.evaluate(env,l+2)
            
        #~ print ' '*l +'operator after re-eval:',str(type(operator))
        #~ if type(operator) is InstanceType:
            #~ print ' '*l+'operator class after re-eval:',operator.__class__.__name__
        #~ print ' '*l + 'evaluating arguments\n'
        
        #~ operands = self.args
        #~ operands = [arg.evaluate(env,l+2) for arg in self.args]
        operands = []
        for arg in self.args:
            if isinstance(arg, LambdaTerm):
                operand = arg.evaluate(env, l+2)
            elif isinstance(arg, LatentFunction):
                operand = arg
            else:
                operand = Constant(arg).evaluate(env, l+2)
            operands += [operand]
        
        #~ print ' '*l +'arguments:\n'
        #~ for a in operands:
            #~ print indent(str(a),l+2) + '\n'
            
        associations = dict(zip(operator.abstraction.params, operands))
        
        #~ print ' '*l + 'new environment:\n'
        
        new_env = operator.env.extend(associations)
        
        #~ print indent(str(new_env), l+2)
        #~ print ' '*l + 'evaluating body:\n'
        if type(operator) is FunctionType:
            new_env['operator'] = operator
            new_env['operands'] = operands
            return eval('operator(*operands)', {}, new_env)
            
        if len(operator.abstraction.params) != len(self.args):
            raise ExecutionError('Wrong number of parameters. I was expecting: '+str(operator.abstraction.params)+' but received: '+str(self.args))
            
        result = operator.abstraction.body.evaluate(new_env,l+2)
        
        #~ print ' '*l + 'final result:'
        #~ print indent(str(result), l+2)
        #~ print ' '*l + 'final result type:',str(type(result))
        #~ if type(result) is InstanceType:
            #~ print ' '*l + 'final result class:',result.__class__.__name__
        #~ print
        return result
    
    def __str__(self):
        args = ', '.join([str(arg) for arg in self.args])
        func = str(self.function) 
        if isinstance(self.function, LambdaAbstraction) \
        or isinstance(self.function, Application):
            func = "(" + func + ")"
        return func + "(" + args + ")"
        #~ return "App(\n" \
            #~ + "  function: \n" \
            #~ + indent(str(self.function), 4) + "\n" \
            #~ + "  arguments: \n  [\n" \
            #~ + indent(',\n'.join([str(arg) for arg in self.args]), 4) + "\n" \
            #~ + "  ]\n)\n"

    """
    Old methods
    """
    def beta_deep_reduce(self, symbol_table={}):
        new_subst = zip(self.function.params, self.args)
        return self.function.body.deep_msubst(new_subst, symbol_table)
        
    def beta_eval(self, symbol_table={}):
        #~ if not isinstance(self.function, LambdaAbstraction):
            #~ raise ReductionError('Operaror is not lambda abstraction.')
        new_subst = zip(self.function.params, self.args)
        return self.function.body.eval_msubst(new_subst, symbol_table)
    
    def deep_subst(self, variable, term):
        return Application(self.function.deep_subst(variable, term), \
                            [arg.deep_subst(variable, term) for arg in self.args])
    
    def old_step(self):
        if isinstance(self.function, Constant):
            return self
        elif isinstance(self.function, Variable):
            return self
        elif isinstance(self.function, LambdaAbstraction):
            return self.beta_eval()
        elif isinstance(self.function, Application):
            return Application(self.function.step(), self.args)
    

class SyntaxError(Exception):
    pass


class ExecutionError(Exception):
    pass


class ReductionError(Exception):
    pass


Const = Constant
Var = Variable
Abs = LambdaAbstraction
App = Application


class LatentFunction:

    def __init__(self, abstraction, env):
        self.abstraction = abstraction
        self.env = env
        
    def __str__(self):
        return str(self.abstraction)
        #~ return "LatentFunction(\n" \
            #~ + "  function: \n" \
            #~ + indent(str(self.abstraction),4) +'\n' \
            #~ + "  environment: \n" \
            #~ + indent(str(self.env), 4) +'\n'\
            #~ + ")\n"


name_counter = 0

def fresh_name():
    global name_counter
    name_counter += 1
    return 'z'+str(name_counter)
